##########################################################################
# xmlexcel - convert XML into a comma-separated CVS file
#            suitable for importing as a spreadsheet
#
# Tony Garland - contact@SpiritAndTruth.org
#
# Anyone may freely redistribute and use this program.
#
##########################################################################
import os.path
import sys
import re
import string
import time
import codecs
from xml.sax import make_parser
from xml.sax.handler import feature_namespaces
from xml.sax import saxutils

HELP='''
Usage: xmlexcel [flags] file1 [file2 . . . ]
-----

    This program scans any ASCII file containing XML elements and converts the
    contents into a TXT file suitable for importing to a spreadsheet program
    such as Excel. The user specifies an optional element name by which
    individual records are denoted within the XML.

    Subelements within each record are converted into individual columns 
    in the resulting spreadsheet with one record per line. When a subelement
    occurs multiple times within a single record, all the values of the
    subelement are merged into a single column.

    For example:

        <holiday>
            <name>Christmas</name>
            <color>red</color>
            <color>green</color>
            <month>December</month>
        </holiday>
        <holiday>
            <name>Independence Day</name>
            <color>red</color>
            <color>white</color>
            <color>blue</color>
            <month>July</month>
        </holiday>
        ...

    Converts as:

        -------------------------------------------------+
        | name             |  color           | month    |
        -------------------------------------------------+
        | Christmas        |  red green   `   | December |
        | Independence Day |  red white blue  | July     |
        ...

Flags:
-----

-?              Print help information.

-help           Print help information.

-element name   Specify the name of the XML element which spans 
                each record. For example, if the XML file 
                presents each record as follows:

                    <item>
                        <author>John Q. Public</author>
                        <title>My Life</title>
                        <date>2004</date>
                    </item>
                    <item>
                        <author>Fred Fisher</author>
                        <title>Life in the Sun</title>
                        <date>1924</date>
                    </item>

                Then specify:

                    -element item

                If no element is specified, then a the record
                separator is assumed to be "record" as if you
                had specified:

                    -element record

Running in Windows:
------------------

    The program is available as an XMLEXCEL.EXE and may be run in Windows as
    follows:

    1. Unzip the downloaded distribution into a folder. We'll call this
       FOLDER1.
    2. Open explorer in FOLDER1.
    3. Open another copy of explorer where your XML file is that you
       want to convert. We'll call this FOLDER2.
    4. Drag the XML file from FOLDER2 and drop it over the
       XMLEXCEL.EXE file in FOLDER1.
    5. The XML file will be converted and a file by the same name, but
       with a TXT extension will appear in FOLDER2.
    6. Open the newly-created file with your speadsheet program (e.g., Excel).

Written by Tony Garland, contact@SpiritAndTruth.org
'''
VERSION='0.1'
HISTORY='''

0.1     Initial version produced for use by Todd Bolen of
        www.BiblePlaces.com.
'''

def printHelp() : 
    print 'Version:', VERSION
    print HELP
    print HISTORY

class docHandler( saxutils.DefaultHandler ) :
    # Document Handler derives from XML parser class for callbacks.
    def __init__( self, outFile, recordElement, args ) :
        self.m_outFile = outFile
        self.m_recordElement = recordElement
        self.m_args = args
        self.m_inRecord = 0
        self.m_fieldContent = {}
        self.m_fieldNames = []
        self.m_fieldName = ''
        self.m_inElement = 0
        self.m_records = []
        self.m_firstElementName = ''
        self.m_excelFieldSeparator = '\t'

    def startElement( self, name, attrs ) :
        # Callback from XML parser at the beginning of each XML element.

        # If haven't seen any element yet,
        # keep track of first one so we know when we are done.
        if not self.m_firstElementName : 
            self.m_firstElementName = name

        # If this is the element which begins each record.
        if name == self.m_recordElement : 
            
            # Get ready to parse the next record.
            self.m_inRecord = 1
            self.m_fieldContent = {}
            self.m_fieldNames = []

        # If we are currently within a record.
        if self.m_inRecord :

            # Keep track of all the different element names nested
            # within this record.
            self.m_fieldNames.append( name )

            # If this element name has previously been seen.
            if self.m_fieldContent.has_key ( name ) :

                # We'll be appending more information to the existing
                # element.
                self.m_fieldContent[ name ] += ' '

            # Else first time this element has been seen.
            else:

                # Create new empty entry ready to append info to.
                self.m_fieldContent[ name ] = ''

        # We are now within an element.
        self.m_inElement = 1

    def characters( self, ch ):
        # Callback from XML parser when user characters (outside of
        # XML element tags) is encountered.

        # If we are within a record and within a subelement.
        if self.m_inRecord and self.m_inElement :

            # Clean up end-of-line.
            ch = re.sub( '[\r\n]', r'', ch )

            # Ignore empty fields completely.
            ch = re.sub( '^\s+$', r'', ch )

            # Compress multiple spaces from XML into a single space.
            ch = re.sub( '\s+', r' ', ch )

            # Append the user characters to the appropriate element
            # name information. (The characters are associated with the
            # most recently-encountered element name.)
            # Elements with duplicate names have their information
            # merged into the first occurrence.
            self.m_fieldContent[ self.m_fieldNames[ -1 ] ] += ch

    def endElement( self, name ):
        # Callback from XML parser when a closing element is found.

        # If this is the close of the record.
        if name == self.m_recordElement : 

            # No longer in a record.
            self.m_inRecord = 0

            # Make a dictionary containing the information for
            # each element keyed by the element name.
            # (Duplicate named elements have had their contents merged.)
            prevName = ''
            infoByName = {}
            for name in self.m_fieldNames :

                # Drop duplicated names.
                if name == prevName : continue

                # Get the information associated with that element.
                # (We ignore empty fields.)
                prevName = name
                info = self.m_fieldContent[ name ]
                if info : infoByName[ name ] = info

            # Keep track of the elements associated with each field.
            self.m_records.append( infoByName )

        # We are no longer in an element.
        self.m_inElement = 0

        # Is this the element which opened the file?
        if name == self.m_firstElementName : 

            # Print the output.
            self.printRecords()

    def printRecords( self ) :
        # Print the output now that the XML file has been read.

        # Print file header info.
        header = ''
        header += 'File automatically generated by %s on %s\n' % ( args[ 0 ], time.ctime() )
        header += 'Command: %s\n' % string.join( args[0 : ], ' ' )
        header += 'Contact: Tony Garland, contact@SpiritAndTruth.org, www.SpiritAndTruth.org\n'
        self.m_outFile.write( '%s\n' % header )

        # Build list of all unique names found in any record.
        allnames = []
        for infoByName in self.m_records :
            for name in infoByName.keys() :
                if not allnames.count( name ) :
                    allnames.append( name )

        # Print heading line.
        self.m_outFile.write( \
            string.join( allnames, self.m_excelFieldSeparator ) \
            )
        self.m_outFile.write( '\n' )

        # For each record, construct the associate information.
        print 'Found %s records' % len( self.m_records )
        for infoByName in self.m_records :
            fields = []

            # Every name ever found in any record.
            for name in allnames :

                # If this record had that name.
                if infoByName.has_key( name ) :

                    # Format the info for import to Excel.
                    info = self.excelFormat( infoByName[ name ] )
                    fields.append( info )

                # Else an empty record (non-empty in some other record).
                else :
                    fields.append( '-' )

            # Write the fields as an individual line in the Excel import file.
            self.m_outFile.write( \
                string.join( fields,  self.m_excelFieldSeparator ) \
                )
            self.m_outFile.write( '\n' )

    def excelFormat( self, text ) :
        # Convert text ready to output in Excel text format.
        # Fields can't contain the separator character itself.
        return re.sub( self.m_excelFieldSeparator, ' ', text )

if __name__ == '__main__':

    args = sys.argv
    if args.count( '-?' ) :
        args.remove( '-?' )
        printHelp()
        sys.exit( 1 )

    if args.count( '-help' ) :
        args.remove( '-help' )
        printHelp()
        sys.exit( 1 )

    element = 'record'
    if args.count( '-element' ) :
        idx = args.index( '-element' )
        args.remove( '-element' )
        element = args[ idx ]
        args.remove( element )
    print 'Using element marker: <%s>' % element

    # Usage.
    if len( args ) < 2 :
        print 'Usage: %s [flags] file1 [file2 . . . ]' % sys.argv[ 0 ]
        printHelp()
        sys.exit( 1 )

    # Get Unicode reader and writer.
    ( unic_encode, unic_decode, unic_reader, unic_writer ) = codecs.lookup( 'utf-8' )

    # Create a parser
    parser = make_parser()

    # Tell the parser we are not interested in XML namespaces
    parser.setFeature(feature_namespaces, 0)

    # For each file passed as input.
    for file in sys.argv[ 1 : ] : 
        print file

        # Create a corresponding output file.
        inFileName = file
        outFileName, ext = os.path.splitext( file )
        outFileName += '.txt'
        inFile = outFile = None
        try :

            # We read and write through unicode translation layer.
            inFile = open( inFileName, 'r' )
            inFileReader = unic_reader( inFile )
            outFile = open( outFileName, 'w' )
            outFileWriter = unic_writer( outFile )
            print '%s -> %s' % ( inFileName, outFileName )

            # Create the document handler
            handler = docHandler( outFileWriter, element, args )

            # Tell the parser to use our handler
            parser.setContentHandler( handler )

            # Parse the file
            parser.parse( inFile )

        finally :
            if inFile: inFile.close()
            if outFile: outFile.close()

